@@ -1,5 +1,5 @@
 #' @title Metropolis-within-Gibbs sampler for spectral inference of a stationary time series using a B-spline prior
-#' @description This function updates the (cubic) B-spline prior using the Whittle likelihood and obtains samples from the pseudo-posterior to infer the spectral density of a stationary time series.
+#' @description This function updates the B-spline prior using the Whittle likelihood and obtains samples from the pseudo-posterior to infer the spectral density of a stationary time series.
 #' @details The function \code{gibbs_bspline} is an implementation of the (serial version of the) MCMC algorithm presented in Edwards et al. (2018).  This algorithm uses a nonparametric B-spline prior to estimate the spectral density of a stationary time series and can be considered a generalisation of the algorithm of Choudhuri et al. (2004), which used the Bernstein polynomial prior.  A Dirichlet process prior is used to find the weights for the B-spline densities used in the finite mixture and a seperate and independent Dirichlet process prior used to place knots.  The algorithm therefore allows for a data-driven choice of the number of knots/mixtures and their locations.
 #' @param data numeric vector
 #' @param Ntotal total number of iterations to run the Markov chain
@@ -12,41 +12,42 @@
 #' @param MH Dirichlet process base measure constant for knot placements of B-spline densities (> 0)
 #' @param H0.alpha,H0.beta parameters of Beta base measure of Dirichlet process for knot placements of B-spline densities (default is Uniform[0, 1])
 #' @param LH truncation parameter of Dirichlet process in stick breaking representation for knot placements of B-spline densities
-#' @param tau.alpha,tau.beta prior parameters for tau (Inverse Gamma)
+#' @param tau.alpha,tau.beta prior parameters for tau (Inverse-Gamma)
 #' @param kmax upper bound for number of B-spline densities in mixture
-#' @param k1 starting value for k.  If k1 = NA then a random starting value between 5 and kmax is selected.
-#' @return A list containing the following components:
+#' @param k1 starting value for k.  If \code{k1} = NA then a random starting value between \code{degree} + 2 and \code{kmax} is selected
+#' @param degree positive integer specifying the degree of the B-spline densities (default is 3) 
+#' @return A list with S3 class 'psd' containing the following components:
 #'    \item{psd.median,psd.mean}{psd estimates: (pointwise) posterior median and mean}
-#'    \item{psd.p05,psd.p95}{pointwise credibility interval}
-#'    \item{psd.u05,psd.u95}{uniform credibility interval}
+#'    \item{psd.p05,psd.p95}{90\% pointwise credibility interval}
+#'    \item{psd.u05,psd.u95}{90\% uniform credibility interval}
 #'    \item{k,tau,V,Z,U,X}{posterior traces of model parameters}
 #'    \item{knots.trace}{trace of knot placements}
 #'    \item{ll.trace}{trace of log likelihood}
-#' @references Edwards, M. C., Meyer, R., and Christensen, N. (2018), Bayesian nonparametric spectral density estimation using B-spline priors, \emph{Statistics and Computing}, https://doi.org/10.1007/s11222-017-9796-9.
+#'    \item{pdgrm}{periodogram}
+#'    \item{n}{integer length of input time series}
+#' @seealso \link{plot.psd}
+#' @references Edwards, M. C., Meyer, R., and Christensen, N. (2018), Bayesian nonparametric spectral density estimation using B-spline priors, \emph{Statistics and Computing}, <https://doi.org/10.1007/s11222-017-9796-9>.
 #' 
 #' Choudhuri, N., Ghosal, S., and Roy, A. (2004), Bayesian estimation of the spectral density of a time series, \emph{Journal of the American Statistical Association}, 99(468):1050--1059.
 #' 
 #' @examples 
 #' \dontrun{
 #' 
+#' set.seed(123456)
+#' 
 #' # Generate AR(1) data with rho = 0.9
 #' n = 128
 #' data = arima.sim(n, model = list(ar = 0.9))
 #' data = data - mean(data)
 #'
 #' # Run MCMC (may take some time)
-#' mcmc = gibbs_bspline(data, Ntotal = 4000, burnin = 2000, thin = 1)
+#' mcmc = gibbs_bspline(data, 10000, 5000)
 #'
-#' # Compare estimate with true PSD
 #' require(beyondWhittle)  # For psd_arma() function
-#' freq = 2 * pi / n * (1:(n / 2 + 1) - 1)[-c(1, n / 2 + 1)]
-#' psd.true <- psd_arma(freq, ar = 0.9, ma = numeric(0), sigma2 = 1)
-#' plot(x = freq, y = psd.true, col = 2, type = "l", xlab = "Frequency", ylab = "PSD")
-#' lines(x = freq, y = mcmc$psd.median, type = "l")
-#' lines(x = freq, y = mcmc$psd.p05, type = "l", lty = 2)
-#' lines(x = freq, y = mcmc$psd.p95, type = "l", lty = 2)
-#' legend(x = "topright", legend = c("true psd", "pointwise median", "pointwise CI"), 
-#' lty = c(1, 1, 2), col = c(2, 1, 1))
+#' freq = 2 * pi / n * (1:(n / 2 + 1) - 1)[-c(1, n / 2 + 1)]  # Remove first and last frequency
+#' psd.true = psd_arma(freq, ar = 0.9, ma = numeric(0), sigma2 = 1)  # True PSD
+#' plot(mcmc)  # Plot log PSD (see documentation of plot.psd)
+#' lines(freq, log(psd.true), col = 2, lty = 3, lwd = 2)  # Overlay true PSD
 #' }
 #' @importFrom Rcpp evalCpp
 #' @useDynLib bsplinePsd, .registration = TRUE
@@ -67,11 +68,18 @@ gibbs_bspline <- function(data,
                           tau.alpha = 0.001,
                           tau.beta = 0.001,
                           kmax = 100,
-                          k1 = 20) {
+                          k1 = 20,
+                          degree = 3) {
   
   n <- length(data)
   
-  if (n %% 2 != 0) stop("this version of bsplinePsd must have n even")
+  # Which boundary frequencies to remove from likelihood computation and tau sample
+  if (n %% 2) {  # Odd length time series
+    bFreq <- 1  # Remove first
+  } 
+  else {  # Even length time series
+    bFreq <- c(1, n)  # Remove first and last
+  }
   
   # Tolerance for mean centering
   tol <- 1e-4
@@ -79,13 +87,12 @@ gibbs_bspline <- function(data,
   # Mean center
   if (abs(mean(data)) > tol) {
     data <- data - mean(data)
-    warning("Data has been mean-centred")
+    warning("data has been mean-centered")
   }
   
   if (burnin >= Ntotal) stop("burnin must be less than Ntotal")
-  if (any(c(Ntotal, burnin, thin) %% 1 != 0 || any(c(Ntotal, burnin, thin) < 0))) 
   if (any(c(MG, MH, G0.alpha, G0.beta, H0.alpha, H0.beta, tau.alpha, tau.beta, k.theta) <= 0)) stop("MG, MH, G0.alpha, G0.beta, H0.alpha, H0.beta, tau.alpha, tau.beta, and k.theta must be strictly positive")
-  if (any(c(Ntotal, thin, kmax, LG, LH) %% 1 != 0) || any(c(Ntotal, thin, kmax, LG, LH) <= 0)) stop("Ntotal, kmax, LG, and LH must be strictly positive integers")
+  if (any(c(Ntotal, thin, kmax, LG, LH) %% 1 != 0) || any(c(Ntotal, thin, kmax, LG, LH) <= 0)) stop("Ntotal, thin, kmax, LG, and LH must be strictly positive integers")
   if ((burnin %% 1 != 0) || (burnin < 0)) stop("burnin must be a non-negative integer")
 
   FZ <- fast_ft(data)  # FFT data to frequency domain.  NOTE: Must be mean-centred.
@@ -111,16 +118,16 @@ gibbs_bspline <- function(data,
   Z[, 1] <- stats::rbeta(LH + 1, H0.alpha, H0.beta)  # G0.alpha = G0.beta = 1 gives U[0,1]
   
   if (is.na(k1)) {  # If k1 is NA, user does not specify starting value for k
-    k[1] = sample(5:kmax, 1)  # Need at least k = 5 for cubic B-splines
+    k[1] = sample((degree + 2):kmax, 1)  # Need at least k = 5, 4, 3 for cubic, quadratic, linear B-splines respectively
   }
   else {  # User specified starting value for k
-    if ((k1 < 5) || (k1 > kmax)) stop("k must be at least 5 and no more than kmax")  
+    if ((k1 < (degree + 2)) || (k1 > kmax)) stop("k1 must be at least degree + 2 and no more than kmax") 
     k[1] <- k1
   }
 
   # Store log likelihood
   ll.trace <- rep(NA, Ntotal)
-  ll.trace[1] <- llike(omega, FZ, k[1], V[, 1], W[, 1], U[, 1], Z[, 1], tau[1], pdgrm, recompute = TRUE)$llike
+  ll.trace[1] <- llike(omega, FZ, k[1], V[, 1], W[, 1], U[, 1], Z[, 1], tau[1], pdgrm, degree, recompute = TRUE)$llike
   
   # Metropolis proposal parameters for V, U, W, Z.
   epsG <- seq(1, LG + 1) / (seq(1, LG + 1) + 2 * sqrt(n))  
@@ -155,6 +162,7 @@ gibbs_bspline <- function(data,
                      tau.alpha,
                      tau.beta,
                      pdgrm,
+                     degree,
                      recompute = TRUE)
     
     f.store <- lp.list$lp
@@ -171,7 +179,7 @@ gibbs_bspline <- function(data,
       jump <- round(stats::rt(1, 1))  # Bold proposal - discrete Cauchy
     }
     k.star <- k[i] + jump
-    while (k.star < 5 || k.star > kmax) {  # A bit hacky to ensure k doesn't go out of bounds
+    while (k.star < (degree + 2) || k.star > kmax) {  # A bit hacky to ensure k doesn't go out of bounds
       if (bold < 0.75) {
         jump <- sample(-1:1, 1, prob = rep(1 / 3, 3))  # Ordinary proposal
       }
@@ -200,6 +208,7 @@ gibbs_bspline <- function(data,
                      tau.alpha,
                      tau.beta,
                      pdgrm,
+                     degree,
                      recompute = TRUE)   
     
     f.k.star <- lp.list$lp
@@ -258,6 +267,7 @@ gibbs_bspline <- function(data,
                        tau.alpha,
                        tau.beta,
                        pdgrm,
+                       degree,
                        recompute = FALSE,
                        db.list)  # Do not need to recompute B-splines for these parameters    
       
@@ -317,6 +327,7 @@ gibbs_bspline <- function(data,
                        tau.alpha,
                        tau.beta,
                        pdgrm,
+                       degree,
                        recompute = FALSE,
                        db.list)  # Do not need to recompute B-splines for these parameters  
       
@@ -378,6 +389,7 @@ gibbs_bspline <- function(data,
                        tau.alpha,
                        tau.beta,
                        pdgrm,
+                       degree,
                        recompute = TRUE) 
       
       f.U.star <- lp.list$lp
@@ -437,6 +449,7 @@ gibbs_bspline <- function(data,
                        tau.alpha,
                        tau.beta,
                        pdgrm,
+                       degree,
                        recompute = TRUE)  
       
       f.Z.star <- lp.list$lp
@@ -464,16 +477,18 @@ gibbs_bspline <- function(data,
     # Step 6: Directly sample tau from conjugate Inverse-Gamma density
     #####
     q.psd <- qpsd(omega, k[i + 1], V[, i + 1], W[, i + 1], U[, i + 1], Z[, i + 1], 
-                  recompute = FALSE, db.list)$psd
-    m <- n - 2
-    q <- rep(NA, m)
-    q[1] <- q.psd[1]
-    q[m] <- q.psd[length(q.psd)]
-    q[2 * 1:(m / 2 - 1)] <- q[2 * 1:(m / 2 - 1) + 1] <- q.psd[1:(m / 2 - 1) + 1]
+                  degree, recompute = FALSE, db.list)$psd
+    q <- unrollPsd(q.psd, n)
     
-    # Note the (n - 2) here - we remove the first and last terms
-    tau[i + 1] <- 1 / stats::rgamma(1, tau.alpha + (n - 2) / 2, 
-                                    tau.beta + sum(pdgrm[2:(n - 1)] / q) / (2 * pi) / 2)
+    # Note: (n - 1) and (n - 2) here.  Remove the first and last terms for even and first for odd
+    if (n %% 2) {  # Odd length series
+      tau[i + 1] <- 1 / stats::rgamma(1, tau.alpha + (n - 1) / 2, 
+                                      tau.beta + sum(pdgrm[-bFreq] / q[-bFreq]) / (2 * pi) / 2)
+    }
+    else {  # Even length series
+      tau[i + 1] <- 1 / stats::rgamma(1, tau.alpha + (n - 2) / 2, 
+                                      tau.beta + sum(pdgrm[-bFreq] / q[-bFreq]) / (2 * pi) / 2)
+    }
     #####
     # End: Step 6
     #####
@@ -482,7 +497,7 @@ gibbs_bspline <- function(data,
     # Step 7: Compute log likelihood
     #####
     ll.trace[i + 1] <- llike(omega, FZ, k[i + 1], V[, i + 1], W[, i + 1], U[, i + 1], Z[, i + 1], tau[i + 1], pdgrm,
-                             recompute = FALSE, db.list)$llike
+                             degree, recompute = FALSE, db.list)$llike
     #####
     # End: Step 7
     #####
@@ -499,47 +514,72 @@ gibbs_bspline <- function(data,
   Z <- Z[, keep]
   ll.trace <- ll.trace[keep]
   
-  fpsd.sample <- log.fpsd.sample <- matrix(NA, nrow = length(omega) - 2, ncol = length(keep))
+  fpsd.sample <- log.fpsd.sample <- matrix(NA, nrow = length(omega), ncol = length(keep))
   knots.trace <- matrix(NA, nrow = kmax, ncol = length(keep))
   
   # Store PSDs
   for (isample in 1:length(keep)) {
     q.psd <- qpsd(omega, k[isample], V[, isample], W[, isample], U[, isample], Z[, isample], 
-                  recompute = TRUE)
+                  degree, recompute = TRUE)
     fpsd.sample[, isample] <- tau[isample] * q.psd$psd
     knots.trace[1:length(q.psd$knots), isample] <- q.psd$knots
     log.fpsd.sample[, isample] <- logfuller(fpsd.sample[, isample])  # Create transformed version
   }
   
+  #ci.l = (1 - ci) / 2
+  #ci.u = 1 - ci.l
+  
   # Compute point estimates and 90% Pointwise CIs
   psd.median <- apply(fpsd.sample, 1, stats::median)
   psd.mean <- apply(fpsd.sample, 1, mean)
   psd.p05 <- apply(fpsd.sample, 1, stats::quantile, probs=0.05)
   psd.p95 <- apply(fpsd.sample, 1, stats::quantile, probs=0.95)
+  #psd.p.lower <- apply(fpsd.sample, 1, stats::quantile, probs=ci.l)
+  #psd.p.upper <- apply(fpsd.sample, 1, stats::quantile, probs=ci.u)
   
   # Transformed versions of these for uniform CI construction
   log.fpsd.s <- apply(log.fpsd.sample, 1, stats::median)
   log.fpsd.mad <- apply(log.fpsd.sample, 1, stats::mad)
   log.fpsd.help <- apply(log.fpsd.sample, 1, uniformmax)
   log.Cvalue <- stats::quantile(log.fpsd.help, 0.9)
+  #log.Cvalue <- stats::quantile(log.fpsd.help, ci)
   
   # Compute Uniform CIs
   psd.u95 <- exp(log.fpsd.s + log.Cvalue * log.fpsd.mad)
   psd.u05 <- exp(log.fpsd.s - log.Cvalue * log.fpsd.mad)
   
-  return(list(psd.median = psd.median,
-              psd.mean = psd.mean,
-              psd.p05 = psd.p05,
-              psd.p95 = psd.p95,
-              psd.u05 = psd.u05,
-              psd.u95 = psd.u95,
-              k = k,
-              tau = tau,
-              V = V,
-              Z = W,  # W here is called Z in paper
-              U = U,
-              X = Z,  # Z here is called X in paper
-              knots.trace = knots.trace,
-              ll.trace = ll.trace))
+  # Compute periodogram
+  N = length(psd.median)  # N = (n + 1) / 2 (ODD) or N = n / 2 + 1 (EVEN)
+  pdgrm = (abs(stats::fft(data)) ^ 2 / (2 * pi * n))[1:N]
+  # kappa = rep(2, N)  # Open kappa object.  Most multiply by 2 but ends not.
+  # if (n %% 2) {  # Odd length time series
+  #   kappa[1] = 1  # Zero frequency
+  # } 
+  # else {  # Even length time series
+  #   kappa[c(1, length(kappa))] = 1  # 0 and Nyquist frequency
+  # }
+  # pdgrm = kappa * (abs(stats::fft(data)) ^ 2 / (2 * pi * n))[1:N]
+  
+  # List to output
+  output = list(psd.median = psd.median,
+                psd.mean = psd.mean,
+                psd.p05 = psd.p05,
+                psd.p95 = psd.p95,
+                psd.u05 = psd.u05,
+                psd.u95 = psd.u95,
+                k = k,
+                tau = tau,
+                V = V,
+                Z = W,  # W here is called Z in paper
+                U = U,
+                X = Z,  # Z here is called X in paper
+                knots.trace = knots.trace,
+                ll.trace = ll.trace,
+                pdgrm = pdgrm,
+                n = n)
+  
+  class(output) = "psd"  # Assign S3 class to object
+  
+  return(output)  # Return output
   
 }  # Close function
